use io;

/* Read instance data. The input files follow the "Archetti" format */
function input() {
    usage = "Usage: localsolver irp.lsp " + 
            "inFileName=inputFile [solFileName=outputFile] [lsTimeLimit=timeLimit]";

    if (inFileName == nil) throw usage;

    readInputIrp();

    // Compute distance matrix
    computeDistanceMatrix();
}

/* Declare the optimization model */
function model() {
    // Quantity of product delivered at each discrete time instant of
    // the planning time horizon to each customer
    delivery[t in 0...horizonLength][i in 0...nbCustomers] <- float(0, capacity);

    // Sequence of customers visited at each discrete time instant of
    // the planning time horizon
    route[t in 0...horizonLength] <- list(nbCustomers);

    for [t in 0...horizonLength] {
        local sequence <- route[t];
        local c <- count(sequence);

        // Customers receive products only if they are visited
        isDelivered[t][i in 0...nbCustomers] <- contains(sequence, i);

        // Distance traveled at instant t
        costRoute[t] <- (c > 0) ? distanceSupplier[sequence[0]] + 
                sum(1...c, i => distance[sequence[i-1]][sequence[i]]) + 
                distanceSupplier[sequence[c-1]] : 0;
    }

    // Stockout constraints at the supplier
    inventorySupplier[0] <- startLevelSupplier;
    for [t in 1...horizonLength+1] {
        inventorySupplier[t] <- inventorySupplier[t-1] 
                - sum[i in 0...nbCustomers](delivery[t-1][i])
                + productionRate_supplier;

        if (t != horizonLength)
            constraint inventorySupplier[t] >= sum[i in 0...nbCustomers](delivery[t][i]);
    }

    // Stockout constraints at the customers
    inventory[i in 0...nbCustomers][0] <- startLevel[i];
    for [i in 0...nbCustomers] {
        for [t in 1...horizonLength+1] {
            inventory[i][t] <- inventory[i][t-1] + delivery[t-1][i] - demandRate[i];
            constraint inventory[i][t] >= 0;
        }
    }
    
    for [t in 0...horizonLength] {
        // Capacity constraints
        constraint sum[i in 0...nbCustomers](delivery[t][i]) <= capacity;
        
        // Maximum level constraints
        for [i in 0...nbCustomers] {
            constraint delivery[t][i] <= maxLevel[i] - inventory[i][t];
            constraint delivery[t][i] <= maxLevel[i] * isDelivered[t][i];
        }
    }

    // Total inventory cost at the supplier
    costInventorySupplier <-
            holdingCostSupplier * sum[t in 0...horizonLength+1](inventorySupplier[t]);

    // Total inventory cost at customers
    costInventory <- sum[i in 0...nbCustomers][t in 0...horizonLength+1](
            holdingCost[i] * inventory[i][t]);

    // Total transportation cost
    costRoute <- sum[t in 0...horizonLength](costRoute[t]);

    // Objective: minimize the sum of all costs
    objective <- costInventorySupplier + costInventory + costRoute;
    minimize(objective);
}

/* Parametrize the solver */
function param() {
    if (lsTimeLimit == nil) lsTimeLimit = 20;
}

/* Write the solution in a file with the following format :
 * - total distance run by the vehicle
 * - the nodes visited at each time step (omitting the start/end at the supplier) */
function output() {
    if (solFileName == nil) return;
    local outfile = io.openWrite(solFileName);

    outfile.println(costRoute.value);
    for [t in 0...horizonLength] {
        for [customer in route[t].value] {
            outfile.print(customer + 1, " ");
        }
        outfile.println();
    }
}

function readInputIrp() {
    inFile = io.openRead(inFileName);

    // General data
    firstLine = inFile.readln().trim().split();
    nbCustomers = firstLine[0].toInt()-1;
    horizonLength = firstLine[1].toInt();
    capacity = firstLine[2].toInt();

    // Supplier data
    secondLine = inFile.readln().trim().split();
    xCoordSupplier = secondLine[1].toDouble();
    yCoordSupplier = secondLine[2].toDouble();
    startLevelSupplier = secondLine[3].toInt();
    productionRate_supplier = secondLine[4].toInt();
    holdingCostSupplier = secondLine[5].toDouble();

    // Customers data
    for [i in 0...nbCustomers] {
        elements = inFile.readln().trim().split();
        xCoord[i] = elements[1].toDouble();
        yCoord[i] = elements[2].toDouble();
        startLevel[i] = elements[3].toInt();
        maxLevel[i] = elements[4].toInt();
        minLevel[i] = elements[5].toInt();
        demandRate[i] = elements[6].toInt();
        holdingCost[i] = elements[7].toDouble();
    }
}

function computeDistanceMatrix() {
    distance[i in 0...nbCustomers][j in 0...nbCustomers] =
            round(sqrt((xCoord[i] - xCoord[j]) * (xCoord[i] - xCoord[j])
            + (yCoord[i] - yCoord[j]) * (yCoord[i] - yCoord[j])));
    distanceSupplier[i in 0...nbCustomers] =
            round(sqrt((xCoord[i] - xCoordSupplier) * (xCoord[i] - xCoordSupplier)
            + (yCoord[i] - yCoordSupplier) * (yCoord[i] - yCoordSupplier)));
}
